"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import os, time, win32api, win32process
from ..types import Event, Logger
from ..abc import Module
from ..constants import Lists
from ..utils.webhook import Webhook

class AntiDLL(Module):
    def __init__(self, webhook: Webhook, logger: Logger, exit: bool, report: bool, event: Event) -> None:
        self.webhook: Webhook = webhook
        self.logger: Logger = logger
        self.exit: bool = exit
        self.report: bool = report
        self.event: Event = event
 
    @property
    def name(self) -> str:
        return "Anti DLL"
    
    @property
    def version(self) -> str:
        return "1.0.0"
     
    def BlockDLLs(self) -> None:
        """Blocks blacklisted DLL's from being injected"""
        while True:
            try:
                time.sleep(1)
                EvidenceOfSandbox = []
                allPids: tuple = win32process.EnumProcesses()
                for pid in allPids:
                    try:
                        hProcess: int = win32api.OpenProcess(0x0410, 0, pid)
                        try:
                            curProcessDLLs: tuple = win32process.EnumProcessModules(
                                hProcess
                            )
                            for dll in curProcessDLLs:
                                dllName: str = str(
                                    win32process.GetModuleFileNameEx(hProcess, dll)
                                ).lower()
                                for sandboxDLL in Lists.BLACKLISTED_DLLS:
                                    if sandboxDLL in dllName:
                                        if dllName not in EvidenceOfSandbox:
                                            EvidenceOfSandbox.append(dllName)
                        finally:
                            win32api.CloseHandle(hProcess)
                    except BaseException:
                        pass
                if EvidenceOfSandbox:
                    self.logger.info(
                        f"The Following DLL's: {EvidenceOfSandbox} Were Found Loaded"
                    )
                    if self.report:
                        self.webhook.Send(
                            f"The following DLL's were discovered loaded in processes running on the users system. DLLS: {EvidenceOfSandbox}",
                            self.name,
                        )
                        self.event.dispatch(
                            "dll_attach",
                            f"The following DLL's were discovered loaded in processes running on the users system. DLLS: {EvidenceOfSandbox}",
                            self.name,
                            {EvidenceOfSandbox},
                            dlls=EvidenceOfSandbox,
                        )
                        self.event.dispatch(
                            "pyguardian_detect",
                            f"The following DLL's were discovered loaded in processes running on the users system. DLLS: {EvidenceOfSandbox}",
                            self.name,
                            {EvidenceOfSandbox},
                            dlls=EvidenceOfSandbox,
                        )
                    if self.exit:
                        os._exit(1)
            except BaseException:
                pass


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import os, ctypes
from ctypes import WinDLL
from ..types import Event, Logger
from ..abc import Module
from ..utils.webhook import Webhook

class AntiAnalysis(Module):
    def __init__(self, webhook: Webhook, logger: Logger, exit: bool, report: bool, event: Event) -> None:
        self.webhook: Webhook = webhook
        self.logger: Logger = logger
        self.exit: bool = exit
        self.report: bool = report
        self.event: Event = event

        self.kernel32: WinDLL = ctypes.windll.kernel32
        self.ntdll: WinDLL = ctypes.windll.ntdll
    
    @property
    def name(self) -> str:
        return "Anti Analysis"
    
    @property
    def version(self) -> str:
        return "1.0.0"
    
    def CheckDebugPrivilege(self) -> None:
        """CHeck debug privileges"""
        hToken = ctypes.c_void_p()
        if (self.ntdll.NtOpenProcessToken(
                self.kernel32.GetCurrentProcess(), 0x0020, ctypes.byref(hToken)
            ) != 0
        ):
            return
    
        privileges = ctypes.create_string_buffer(1024)
        return_length = ctypes.c_ulong()
        if (self.ntdll.NtQueryInformationToken(
                hToken,
                0x03,
                ctypes.byref(privileges),
                ctypes.sizeof(privileges),
                ctypes.byref(return_length),
            ) != 0
        ):
            self.ntdll.NtClose(hToken)
            return
        
        debug_privilege = (ctypes.c_int * (return_length.value // 8)).from_buffer(
            privileges
        )
        for priv in debug_privilege:
            if priv.s_luid.LowPart == 21 and priv.s_attributes & 0x00000002:
                self.ntdll.NtClose(hToken)
                self.logger.info("Debug privilege found enabled")
                if self.report:
                    self.webhook.Send("Debug privilege enabled", self.name)
                    self.event.dispatch("debug_privilege_found", "Debug privilege enabled", self.name)
                    self.event.dispatch("pydefender_detect", "Debug privilege enabled", self.name)
                
                if self.exit:
                    os._exit(1)
        
        self.ntdll.NtClose(hToken)
    
    def HideThreads(self) -> None:
        """Hides the current threads"""
        PID = self.kernel32.GetCurrentProcessId()
        hProcess = self.kernel32.OpenProcess(0x1F0FFF, False, PID)
        if hProcess is None:
            return
        
        TID = self.kernel32.GetCurrentThreadId()
        hThread = self.kernel32.OpenThread(0x1F03FF, False, TID)
        if hThread is None:
            self.kernel32.CloseHandle(hProcess)
            return
        self.ntdll.NtSetInformationThread(
            hThread, 0x11, ctypes.byref((ctypes.c_int(1)), ctypes.sizeof(ctypes.c_int))
        )

        self.kernel32.CloseHandle(hThread)
        self.kernel32.CloseHandle(hProcess)
    
    def CheckDebugObject(self) -> None:
        """Check if debug object handle is found"""
        PID = self.kernel32.GetCurrentProcessId()

        hProcess = self.kernel32.OpenProcess(0x1F0FFF, False, PID)
        if hProcess is None:
            return

        HasDebugObject = (
            self.ntdll.NtQueryInformationProcess(
                hProcess,
                0x1E,
                ctypes.byref(ctypes.c_int(0)),
                ctypes.sizeof(ctypes.c_int),
                ctypes.byref(ctypes.c_int(0)),
            )
            == 0
        )
        if HasDebugObject:
            self.kernel32.CloseHandle(hProcess)
            self.logger.info("Debug object handle has been detected")
            if self.report:
                self.webhook.Send("Debug object handle has been detected", self.name)
                self.event.dispatch("debug_object_handle_found", "Debug object handle has been detected", self.name)
                self.event.dispatch("pydefender_detect", "Debug object handle has been detected", self.name)
            
            if self.exit:
                os._exit(1)
        
    def CheckSEDebugName(self) -> None:
        """Check SE debug name"""
        PID = self.kernel32.GetCurrentProcessId()

        hProcess = self.kernel32.OpenProcess(0x1F0FFF, False, PID)
        if hProcess is None:
            return

        DebugObjectHandle = ctypes.c_void_p()
        if (self.ntdll.NtQueryObject(
                hProcess,
                0x1F,
                ctypes.byref(DebugObjectHandle),
                ctypes.sizeof(DebugObjectHandle),
                None,
            )== 0
        ):
            self.kernel32.CloseHandle(hProcess)
            self.logger.info("Debug object handle has been detected")
            if self.report:
                self.webhook.Send("Debug object handle has been detected", self.name)
                self.event.dispatch("debug_object_handle_found", "Debug object handle has been detected", self.name)
                self.event.dispatch("pydefender_detect", "Debug object handle has been detected", self.name)
            
            if self.exit:
                os._exit((1))
    
    def CheckNtGlobalFlag(self) -> None:
        """Check NT_GLOBAL_FLAG_DEBUGGED"""
        PEB = ctypes.c_int(0)
        self.ntdll.NtQueryInformationProcess.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_void_p,
            ctypes.c_ulong,
            ctypes.POINTER(ctypes.c_ulong),
        ]
        self.ntdll.NtQueryInformationProcess(
            self.kernel32.GetCurrentProcess(),
            0,
            ctypes.byref(PEB),
            ctypes.sizeof(PEB),
            None,
        )
        if (PEB.value & 0x00000001) != 0:
            self.logger.info(
                "NT_GLOBAL_FLAG_DEBUGGED Found in the Process Enviroment Block"
            )
            if self.report:
                self.webhook.Send(
                    "NT_GLOBAL_FLAG_DEBUGGED Found in the Process Environment Block",
                    self.name,
                )
                self.event.dispatch(
                    "nt_global_flag_debugged",
                    "NT_GLOBAL_FLAG_DEBUGGED Found in the Process Environment Block",
                    self.name,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    "NT_GLOBAL_FLAG_DEBUGGED Found in the Process Environment Block",
                    self.name,
                )

            if self.exit:
                os._exit(1)

    def CheckHardwareBreakpoints(self) -> None:
        """Check for existing hardware breakpoints"""
        ThreadContext = ctypes.c_void_p()
        TID = self.kernel32.GetCurrentThreadId()
        hThread = self.kernel32.OpenThread(0x1F03FF, False, TID)
        if hThread is None:
            return

        if not self.kernel32.GetThreadContext(hThread, ctypes.byref(ThreadContext)):
            self.kernel32.CloseHandle(hThread)
            return

        if (ThreadContext.contents.Dr0 != 0
            or ThreadContext.contents.Dr1 != 0
            or ThreadContext.contents.Dr2 != 0
            or ThreadContext.contents.Dr3 != 0
        ):
            self.kernel32.CloseHandle(hThread)
            self.logger.info("Hardware Breakpoints Found Set")
            if self.report:
                self.webhook.Send("Hardware breakpoints found set", self.name)
                self.event.dispatch(
                    "hardware_breakpoint_set",
                    "Hardware Breakpoints Found Set",
                    self.name,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    "Hardware Breakpoints Found Set",
                    self.name,
                )

            if self.exit:
                os._exit(1)
        self.kernel32.CloseHandle(hThread)

    def CheckDebugFilterState(self) -> None:
        """Check Debug Filter State Being !=0"""
        """Check debug filter state being != 0 (not equal to 0)"""
        self.ntdll.DbgSetDebugFilterState(0, 0)

        DebugFilterState = ctypes.c_uint()
        self.ntdll.DbgQueryDebugFilterState(0, ctypes.byref(DebugFilterState))

        if DebugFilterState.value != 0:
            self.logger.info("Debug Filter State is not 0")
            if self.report:
                self.webhook.Send(
                    "Debug filter state `!= 0`, debugging detected", self.name
                )
                self.event.dispatch(
                    "debug_filter_state", "Debug Filter State is not 0", self.name
                )
                self.event.dispatch(
                    "pydefender_detect", "Debug Filter State is not 0", self.name
                )
            if self.exit:
                os._exit(1)

    def CheckPEB(self) -> None:
        """Check PEB For BeingDebugged"""

        class PEB(ctypes.Structure):
            _fields_ = [("BeingDebugged", ctypes.c_byte)]

        process = self.kernel32.GetCurrentProcess()

        peb = PEB()
        self.ntdll.NtQueryInformationProcess(
            process, 0, ctypes.pointer(peb), ctypes.sizeof(peb), None
        )

        if peb.BeingDebugged:
            self.logger.info("Process Being Debugged")
            if self.report:
                self.webhook.Send("Process found being debugged", self.name)
                self.event.dispatch(
                    "peb_being_debugged", "Process Found Being Debugged", self.name
                )
                self.event.dispatch(
                    "pyguardian_detect", "Process Found Being Debugged", self.name
                )
            if self.exit:
                os._exit(1)

    def StartAnalyzing(self) -> None:
        if self.report:
            self.logger.info("Starting Anti Analysis Checks")
        self.CheckDebugPrivilege()
        self.CheckDebugObject()
        self.CheckSEDebugName()
        self.CheckNtGlobalFlag()
        self.CheckHardwareBreakpoints()
        self.CheckDebugFilterState()
        self.CheckPEB()
        if self.report:
            self.logger.info("Finished All Anti Analysis Checks")

        if self.report:
            self.logger.info("Hiding Anti Analysis Threads...")
        self.HideThreads()
        if self.report:
            self.logger.info("Anti Analysis Threads Hidden!")


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
from observable import Observable

class DefenderEvent:
    def __init__(self, event: str, text: str, module: str, **kwargs) -> None:
        self.event: str = event
        self.text: str = text
        self.module:str = module
        self.misc: dict = kwargs

class DefenderObservable:
    def __init__(self) -> None:
        self.obs: Observable = Observable()
    
    def dispatch(self, event: str, text: str, module: str, **kwargs) -> DefenderEvent:
        """
        Triggers an event

        Arguments:
            event (string): The event name
            text (string): The text that was given
            module (string): The name of the module that has been triggered
        
        Returns:
            DefenderEvent
        """
        self.obs.trigger(event, text, module, **kwargs)
        return DefenderEvent(event, text, module, **kwargs)


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import ctypes, os, sys, time, pkg_resources, socket, struct, requests, psutil, win32api
from typing import Literal
from ..types import Event, Logger
from ..abc import Module
from ..constants import UserInfo, Lists
from ..utils.http import hasInternet
from ..utils.webhook import Webhook

class Miscellaneous(Module):
    def __init__(self, webhook: Webhook, logger: Logger, exit: bool, report: bool, event: Event) -> None:
        self.webhook: Webhook = webhook
        self.logger: Logger = logger
        self.exit: bool = exit
        self.report: bool = report
        self.event: Event = event
    
    @property
    def name(self) -> str:
        return self.__class__.__name__
    
    @property
    def version(self) -> str:
        return "1.0.0"
    
    def CheckInternet(self) -> None:
        """
        Checks if the user has a valid interent connection or not

        Returns:
            Nothing
        """
        while True:
            time.sleep(5)
            if hasInternet() is False:
                if self.report:
                    self.logger.info("Check Internet Has Failed")
                if self.exit:
                    os._exit(1)

    def CheckRAM(self) -> None:
        """Checks RAM size, if less than 4gb, report."""
        memory: int = psutil.virtual_memory().total
        if memory <= 4294967296:
            self.logger.info("RAM Check Failed")
            if self.report:
                self.webhook.Send("Less than 4 GB of RAM exists on this system", self.name)
                self.event.dispatch(
                    "ram_check",
                    "Less than 4 GB of RAM exists on this system",
                    self.name,
                    ram=memory,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    "Less than 4 GB of RAM exists on this system",
                    self.name,
                    ram=memory,
                )
            if self.exit:
                os._exit(1)


    def CheckIsDebuggerPresent(self) -> None:
        """Checks IsDebuggerPresent and CheckRemoteDebuggerPresent"""
        isDebuggerPresent = ctypes.windll.kernel32.IsDebuggerPresent()

        if isDebuggerPresent:
            self.logger.send("IsDebuggerPresent Returned True")
            if self.report:
                self.webhook.Send("IsDebuggerPresent Returned True", self.name)
                self.event.dispatch("is_debugger_present", "IsDebuggerPresent Returned True", self.name)
                self.event.dispatch("pydefender_detect", "IsDebuggerPresent Returned True", self.name
)
            if self.exit:
                os._exit(1)

        if (ctypes.windll.kernel32.CheckRemoteDebuggerPresent(
                ctypes.windll.kernel32.GetCurrentProcess(), False
            )!= 0
        ):
            self.logger.send("CheckRemoteDebuggerPresent Returned True")
            if self.report:
                self.webhook.Send("CheckRemoteDebuggerPresent returned `True`",self.name,)
                self.event.dispatch(
                    "check_remote_debugger_present",
                    "CheckRemoteDebuggerPresent Returned True",
                    self.name,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    "CheckRemoteDebuggerPresent Returned True",
                    self.name,
                )
            if self.exit:
                os._exit(1)

    def CheckDiskSize(self) -> None:
        """Check Disk Size"""
        minDiskSizeGB: Literal[50] = 50
        if len(sys.argv) > 1:
            minDiskSizeGB = float(sys.argv[1])
        _, diskSizeBytes, _ = win32api.GetDiskFreeSpaceEx()
        diskSizeGB: int = diskSizeBytes / 1073741824

        if diskSizeGB < minDiskSizeGB:
            self.logger.info("Disk Check Failed")
            if self.report:
                self.webhook.Send(f"The current disk size is `{diskSizeGB}GB`, which is less than the minimum requirement.", self.name)
                self.event.dispatch(
                    "disk_size_check",
                    f"The Current Disk Size Is {diskSizeGB}GB, Which Is Less Than The Minimum",
                    self.name,
                    disk_size=diskSizeGB,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    f"The Current Disk Size Is {diskSizeGB}GB, Which Is Less Than The Minimum",
                    self.name,
                    disk_size=diskSizeGB,
                )
            if self.exit:
                os._exit(1)

    def KillTasks(self) -> None:
        """Kills Common RE Tasks"""
        os.system("taskkill /f /im HTTPDebuggerUI.exe >nul 2>&1")
        os.system("taskkill /f /im HTTPDebuggerSvc.exe >nul 2>&1")
        os.system('taskkill /FI "IMAGENAME eq cheatengine*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq httpdebugger*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq processhacker*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq fiddler*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq wireshark*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq rawshark*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq charles*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq cheatengine*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq ida*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq httpdebugger*" /IM * /F /T >nul 2>&1')
        os.system('taskkill /FI "IMAGENAME eq processhacker*" /IM * /F /T >nul 2>&1')
        os.system("sc stop HTTPDebuggerPro >nul 2>&1")
        os.system("sc stop KProcessHacker3 >nul 2>&1")
        os.system("sc stop KProcessHacker2 >nul 2>&1")
        os.system("sc stop KProcessHacker1 >nul 2>&1")
        os.system("sc stop wireshark >nul 2>&1")
        os.system("sc stop npf >nul 2>&1")
        os.system("sc stop HTTPDebuggerPro >nul 2>&1")
        os.system(
            'cmd.exe /c @RD /S /Q "C:\\Users\\%username%\\AppData\\Local\\Microsoft\\Windows\\INetCache\\IE" >nul 2>&1'
        )

    def CheckPaths(self) -> None:
        """Checks Paths on Computer Against Blacklisted Paths"""
        for path in Lists.BLACKLISTED_PATHS:
            if os.path.exists(path):
                self.logger.info("Blacklisted Path Found")
                if self.report:
                    self.webhook.Send("Blacklisted path found", self.name)
                    self.event.dispatch(
                        "blacklisted_path",
                        "Blacklisted Path Found",
                        module = self.name,
                        path=path,
                    )
                    self.event.dispatch(
                        "pydefender_detect",
                        "Blacklisted Path Found",
                        module = self.name,
                        path=path,
                    )
                if self.exit:
                    os._exit(1)
            else:
                pass

    def CheckImports(self) -> None:
        """Checks Current Installed PyPi Packages For Blacklisted Packages"""
        for package in Lists.BLACKLISTED_IMPORTS:
            try:
                dist = pkg_resources.get_distribution(package)
                if dist:
                    self.logger.info(f"{package} Was Found Installed")
                    if self.report:
                        self.webhook.Send(
                            f"`{package}` was found installed",
                            self.name,
                        )
                        self.event.dispatch(
                            "blacklisted_import",
                            f"{package} Was Found Installed",
                            self.name,
                            dist=dist,
                        )
                        self.event.dispatch(
                            "pydefender_detect",
                            f"{package} Was Found Installed",
                            self.name,
                            dist=dist,
                        )
                    if self.exit:
                        os._exit(1)
                else:
                    pass
            except pkg_resources.DistributionNotFound:
                pass

    def CheckOutPutDebugString(self) -> None:
        """Checks OutPutDebugString"""
        win32api.SetLastError(0)
        win32api.OutputDebugString("PythonProtector Intruding...")
        if win32api.GetLastError() != 0:
            self.logger.info("OutputDebugString Is Not 0")
            if self.report:
                self.webhook.Send("OutputDebugString Not Equal To 0", self.name)
                self.event.dispatch(
                    "output_debug_string", "OutputDebugString Not Equal To 0", self.name
                )
                self.event.dispatch(
                    "pydefender_detect", "OutputDebugString Not Equal To 0", self.name
                )
            if self.exit:
                os._exit(1)

    def CheckIPs(self) -> None:
        """Checks User IP Against Blacklisted List"""
        if UserInfo.IP in Lists.BLACKLISTED_IPS:
            self.logger.info(f"{UserInfo.IP} Is A Blacklisted IP Address")
            if self.report:
                self.webhook.Send(
                    f"`{UserInfo.IP}` is a blacklisted IP address", self.name
                )
                self.event.dispatch(
                    "ip_check",
                    f"{UserInfo.IP} Is A Blacklisted IP Address",
                    self.name,
                    ip=UserInfo.IP,
                )
                self.event.dispatch(
                    "pydefender_detect",
                    f"{UserInfo.IP} Is A Blacklisted IP Address",
                    self.name,
                    ip=UserInfo.IP,
                )
            if self.exit:
                os._exit(1)
        else:
            pass

    def CheckCPUCores(self) -> None:
        """Checks CPU Core Count For Being Less Than 1"""
        if int(psutil.cpu_count()) <= 1:
            self.logger.info("CPU Core Count Is Less Than Or Equal To 1")
            if self.report:
                self.webhook.Send("CPU core count is less than or equal to `1`", self.name)
                self.event.dispatch(
                    "cpu_count" "CPU Core Count Is Less Than Or Equal To 1", self.name
                )
                self.event.dispatch(
                    "pydefender_detect" "CPU Core Count Is Less Than Or Equal To 1",
                    self.name,
                )
            if self.exit:
                os._exit(1)

    def IsUsingProxy(self) -> None:
        """Checks If Proxies Are In Use"""
        headers: dict[str, str] = {"User-Agent": "Mozilla/5.0"}
        response = requests.get("https://www.google.com", headers=headers)
        for header in Lists.PROXY_HEADERS:
            if header in response.headers:
                self.logger.info("Proxy Headers In Use")
                if self.report:
                    self.webhook.Send("Proxy headers are in use", self.name)
                    self.event.dispatch(
                        "proxy_headers",
                        "Proxy Headers Being Used",
                        self.name,
                        header=header,
                    )
                    self.event.dispatch(
                        "pydefender_detect",
                        "Proxy Headers Being Used",
                        self.name,
                        header=header,
                    )
                if self.exit:
                    os._exit(1)

        if UserInfo.IP in Lists.PROXY_IPS:
            self.logger.info("Proxy IP In Use")
            if self.report:
                self.webhook.Send("Proxy IP being used", self.name)
                self.event.dispatch(
                    "proxy_ip", "Proxy IP Being Used", self.name, ip=UserInfo.IP
                )
                self.event.dispatch(
                    "pydefender_detect",
                    "Proxy IP Being Used",
                    self.name,
                    ip=UserInfo.IP,
                )
            if self.exit:
                os._exit(1)

        try:
            _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            _socket.settimeout(5)
            _socket.connect(("check.torproject.org", 9050))
            _socket.send(b"GET / HTTP/1.1\r\nHost: check.torproject.org\r\n\r\n")
            data: bytes = _socket.recv(1024)
            if "Congratulations" in data.decode():
                self.logger.info("Tor Network Detected")
                if self.report:
                    self.webhook.Send("Tor network in use", self.name)
                    self.event.dispatch("tor_network", "Tor Network In Use", self.name)
                    self.event.dispatch(
                        "pydefender_detect", "Tor Network In Use", self.name
                    )
                if self.exit:
                    os._exit(1)
        except Exception:
            pass

        try:
            IP = struct.unpack("!I", socket.inet_aton(UserInfo.IP))[0]
            if IP >> 24 in [0, 10, 100, 127, 169, 172, 192]:
                self.logger.info("Transparent Proxies Detected")
                if self.report:
                    self.webhook.Send("Transparent proxies detected", self.name)
                    self.event.dispatch(
                        "transparent_proxies", "Transparent Proxies Detected", self.name
                    )
                    self.event.dispatch(
                        "pydefender_detect", "Transparent Proxies Detected", self.name
                    )
                if self.exit:
                    os._exit(1)
        except Exception:
            pass

    def StartChecks(self) -> None:
        if self.report:
            self.logger.info("Starting Miscellaneous Checks")
        self.CheckImports()
        self.CheckPaths()
        self.CheckIPs()
        self.CheckCPUCores()
        self.CheckRAM()
        self.CheckIsDebuggerPresent()
        self.CheckOutPutDebugString()
        self.CheckDiskSize()
        self.KillTasks()
        self.IsUsingProxy()
        if self.report:
            self.logger.info("Finished Miscellaneous Checks")

        self.CheckInternet()


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import ctypes, win32api
from ctypes import WinDLL
from ctypes import wintypes
from ..types import Event, Logger
from ..abc import Module
from ..utils.webhook import Webhook

class AntiDump(Module):
    def __init__(self, webhook: Webhook, logger: Logger, exit: bool, report: bool, event: Event) -> None:
        self.webhook: Webhook = webhook
        self.logger: Logger = logger
        self.exit: bool = exit
        self.report: bool = report
        self.event: Event = event

        self.kernel32: WinDLL = ctypes.windll.kernel32
        self.ntdll: WinDLL = ctypes.windll.ntdll
 
    @property
    def name(self) -> str:
        return "Anti DLL"
    
    @property
    def version(self) -> str:
        return "1.0.0"

    def ErasePEHeaderFromMemory(self) -> None:
        """
        Erases PE Header from memory
        """
        oldProtect = wintypes.DWORD(0)

        baseAddress = ctypes.c_int(win32api.GetModuleHandle(None))

        self.kernel32.VirtualProtect(
            ctypes.pointer(baseAddress), 4096, 0x04, ctypes.pointer(oldProtect)
        )
        ctypes.memset(ctypes.pointer(baseAddress), 4096, ctypes.sizeof(baseAddress))
        self.event.dispatch(
            "pe_header_erased", "PE Header Erased From Memory", self.name
        )

    def StartChecks(self) -> None:
        if self.report:
            self.logger.info("Starting Anti Dump")
        if self.report:
            self.logger.info("Erasing PE Header From Memory")
        self.ErasePEHeaderFromMemory()
        if self.report:
            self.logger.info("PE Header Erased From Memory")
            self.logger.info("Finished Anti Dump")


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import os, win32gui, time, psutil
from win32process import GetWindowThreadProcessId
from ..types import Event, Logger
from ..abc import Module
from ..constants import Lists
from ..utils.webhook import Webhook

class AntiProcess(Module):
    def __init__(self, webhook: Webhook, logger: Logger, exit: bool, report: bool, event: Event) -> None:
        self.webhook: Webhook = webhook
        self.logger: Logger = logger
        self.exit: bool = exit
        self.report: bool = report
        self.event: Event = event
 
    @property
    def name(self) -> str:
        return "Anti Process"
    
    @property
    def version(self) -> str:
        return "1.0.0"

    def CheckProcessList(self) -> None:
        """
        Checks the running process list for any blacklisted applications / programs
        """
        while True:
            try:
                time.sleep(0.7)
                for process in psutil.process_iter():
                    if any(
                        process_name in process.name().lower()
                        for process_name in Lists.BLACKLISTED_PROGRAMS
                    ):
                        try:
                            if self.report:
                                self.logger.info(f"{process.name} Process Was Running")
                                self.webhook.Send(
                                    f"`{process.name()}` was detected running on the system.",
                                    self.name,
                                )
                                self.event.dispatch(
                                    "process_running",
                                    f"{process.name()} was detected running on the system.",
                                    self.name,
                                    process=process,
                                )
                                self.event.dispatch(
                                    "pydefender_detect",
                                    f"{process.name()} was detected running on the system.",
                                    self.name,
                                    process=process,
                                )
                            process.kill()
                            if self.exit:
                                os._exit(1)
                        except (psutil.NoSuchProcess, psutil.AccessDenied):
                            pass
            except BaseException:
                pass

    def CheckWindowNames(self) -> None:
        """Checks window names that are in the blacklisted list"""
        def winEnumHandler(hwnd, ctx) -> None:
            if win32gui.GetWindowText(hwnd).lower() in Lists.BLACKLISTED_WINDOW_NAMES:
                pid: tuple[int, int] = GetWindowThreadProcessId(hwnd)
                if isinstance(pid, int):
                    try:
                        psutil.Process(pid).terminate()
                    except BaseException:
                        pass
                else:
                    for process in pid:
                        try:
                            psutil.Process(process).terminate()
                        except BaseException:
                            pass
                self.logger.info(f"{win32gui.GetWindowText(hwnd)} Found")
                if self.report:
                    self.webhook.Send(
                        f"Debugger {win32gui.GetWindowText(hwnd)}", self.name
                    )
                    self.event.dispatch(
                        "window_name_detected",
                        f"Debugger {win32gui.GetWindowText(hwnd)} Found Open",
                        self.name,
                        window_name=win32gui.GetWindowText(hwnd),
                    )
                    self.event.dispatch(
                        "pydefender_detect",
                        f"Debugger {win32gui.GetWindowText(hwnd)} Found Open",
                        self.name,
                        window_name=win32gui.GetWindowText(hwnd),
                    )
                if self.exit:
                    os._exit(1)

        while True:
            win32gui.EnumWindows(winEnumHandler, None)



"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""

import os
import platform
import humanize
import psutil
import sys
import cpuinfo
import datetime
from pathlib import Path
from threading import Thread
from typing import Dict, Union, List, Tuple, Any, Optional
from command_runner.elevate import is_admin
from loguru import logger
from .constants import DefenderInfo, LoggingInfo, UserInfo, Valid
from .modules.AntiProcess import AntiProcess
from .modules.AntiVM import AntiVM
from .modules.AntiDll import AntiDLL
from .modules.AntiAnalysis import AntiAnalysis
from .modules.AntiDump import AntiDump
from .modules.Miscellaneous import Miscellaneous
from .utils.events import DefenderObservable
from .utils.webhook import Webhook
from .utils.exceptions import ModulesNotValid, DetectionsNotValid, LogsPathEmpty

class PyDefender:
    def __init__(
        self, 
        debug: Optional[bool], 
        modules: List[str], 
        webhook_url: Optional[str], 
        on_detect: Optional[List[str]], 
        logs_path: Optional[Union[Path, str]] = None
    ) -> None:
        # -- Validate Input Modules -- #
        self.modules: List[str] = modules
        _modules_valid: bool = Valid.Modules.issuperset(self.modules)
        if not _modules_valid:
            raise ModulesNotValid(
                "List of modules provided does not match, consider checking valid modules."
            )
        
        # -- Validate Input Detections -- #
        self.detections: List[str] = on_detect
        _detections_valid: bool = Valid.Detections.issuperset(self.detections)
        if not _detections_valid:
            raise DetectionsNotValid(
                "List of detection methods does not match, consider checking valid detections."
            )
        
        # -- Initialize Logging -- #
        self.debug: bool | None = debug
        self.logs_path: Path | str | None = logs_path
        self.logger = logger

        if self.debug and not self.logs_path:
            raise LogsPathEmpty("Debug Enabled, but there is no log path provided.")

        if self.logs_path and not self.debug:
            raise RuntimeWarning("Logs Path Was Provided But Debug Was Disabled.")
    
        if self.debug and self.logs_path:
            LOGGING_CONFIG: Dict = {
                "handlers": [{ "sink": self.logs_path, "format": LoggingInfo.encrypted_formatter, "enqueue": True, "rotation": "daily", "mode": "w", "level": "INFO", "serialize": False, "backtrace": False, "catch": False, },],
            }
            self.logger.configure(**LOGGING_CONFIG)

        # -- Initialize Constants -- #
        self.screenshot: bool = bool("Screenshot" in self.detections)
        self.exit: bool = bool("Exit" in self.detections)
        self.report: bool = bool("Report" in self.detections)

        # -- Initialize Events
        self.event: DefenderObservable = DefenderObservable()

        # -- Initialize Webhooks
        self.webhook_url: str = webhook_url

        if self.report and self.webhook_url is None:
            raise RuntimeWarning("Reporting Was Set But No Webhook URL Was Provided.")

        self.webhook: Webhook = Webhook(
            self.webhook_url, self.logs_path, self.screenshot
        )

        # -- Initialize Modules -- #
        self.Miscellaneous: Miscellaneous = Miscellaneous(
            self.webhook, self.logger, self.exit, self.report, self.event
        )
        self.AntiProcess: AntiProcess = AntiProcess(
            self.webhook, self.logger, self.exit, self.report, self.event
        )
        self.AntiDLL: AntiDLL = AntiDLL(
            self.webhook, self.logger, self.exit, self.report, self.event
        )
        self.AntiVM: AntiVM = AntiVM(
            self.webhook, self.logger, self.exit, self.report, self.event
        )
        self.AntiAnalysis: AntiAnalysis = AntiAnalysis(
            self.webhook, self.logger, self.exit, self.report, self.event
        )
        self.AntiDump: AntiDump = AntiDump(
            self.webhook, self.logger, self.exit, self.report, self.event
        )

        # -- Debug Checks -- #
        if self.debug:
            self.logger.enable("PyDefender")
        else:
            self.logger.disable("PyDefender")

    @property
    def version(self) -> str:
        # -- Returns the current PyDefender version -- #
        return DefenderInfo.VERSION

    @property
    def user(self) -> Tuple[str, str, str]:
        # -- Returns a Tuple of user information
        return (UserInfo.PC_NAME, UserInfo.USERNAME, UserInfo.HWID)

    @property
    def ip(self) -> str:
        # -- Returns the current IP Address
        return UserInfo.IP

    @property
    def computer(self) -> Any:
        # -- Returns the current users WMI computer object -- #
        return UserInfo.COMPUTER

    def _run_module_threads(self, debug: bool) -> None:
        # -- Starts threads for each module -- #

        if debug:
            if "Miscellaneous" in self.modules:
                self.logger.info("Starting Miscellaneous Thread")
                Thread(
                    name=self.Miscellaneous.name, target=self.Miscellaneous.StartChecks
                ).start()
                self.logger.info("Miscellaneous Thread Started")
            if "AntiProcess" in self.modules:
                self.logger.info("Starting Anti Process Thread")
                Thread(
                    name="Anti Process List", target=self.AntiProcess.CheckProcessList
                ).start()
                Thread(
                    name="Anti Window Names", target=self.AntiProcess.CheckWindowNames
                ).start()
                self.logger.info("Anti Process Thread Started")
            if "AntiDLL" in self.modules:
                self.logger.info("Starting Anti DLL Thread")
                Thread(name=self.AntiDLL.name, target=self.AntiDLL.BlockDLLs).start()
                self.logger.info("Anti DLL Thread Started")
            if "AntiVM" in self.modules:
                self.logger.info("Starting Anti VM Thread")
                Thread(name=self.AntiVM.name, target=self.AntiVM.StartChecks).start()
                self.logger.info("Anti VM Thread Started")
            if "AntiAnalysis" in self.modules:
                self.logger.info("Starting Anti Analysis Thread")
                Thread(
                    name=self.AntiAnalysis.name, target=self.AntiAnalysis.StartAnalyzing
                ).start()
                self.logger.info("Anti Analysis Thread Started")
            if "AntiDump" in self.modules:
                self.logger.info("Starting Anti Dump Thread")
                Thread(
                    name=self.AntiDump.name, target=self.AntiDump.StartChecks
                ).start()
                self.logger.info("Started Anti Dump Thread")
        else:
            if "Miscellaneous" in self.modules:
                Thread(
                    name=self.Miscellaneous.name, target=self.Miscellaneous.StartChecks
                ).start()
            if "AntiProcess" in self.modules:
                Thread(
                    name="Anti Process List", target=self.AntiProcess.CheckProcessList
                ).start()
                Thread(
                    name="Anti Window Names", target=self.AntiProcess.CheckWindowNames
                ).start()
            if "AntiDLL" in self.modules:
                Thread(name=self.AntiDLL.name, target=self.AntiDLL.BlockDLLs).start()
            if "AntiVM" in self.modules:
                Thread(name=self.AntiVM.name, target=self.AntiVM.StartChecks).start()
            if "AntiAnalysis" in self.modules:
                Thread(
                    name=self.AntiAnalysis.name, target=self.AntiAnalysis.StartAnalyzing
                ).start()
            if "AntiDump" in self.modules:
                Thread(
                    name=self.AntiDump.name, target=self.AntiDump.StartChecks
                ).start()

    def start(self) -> None:
        # -- Main function for PyDefender -- #

        # -- Check if windows platform -- #
        if sys.platform != "win32":
            os._exit(1)

        if platform.python_version_tuple()[1] < "11":
            raise DeprecationWarning("Python is not version 3.11 or above.")

        # -- Start Main Program
        if self.debug:
            self.logger.info("PyDefender Starting")

            self.logger.info(f"Version: {DefenderInfo.VERSION}")
            self.logger.info(f"Current Path: {DefenderInfo.ROOT_PATH}")
            self.logger.info(
                f"Operating System: {platform.uname().system} {platform.uname().release} {platform.win32_edition()} ({platform.architecture(sys.executable)[0]})"
            )
            bt = datetime.datetime.fromtimestamp(psutil.boot_time())
            self.logger.info(
                f"Boot Time: {bt.year}/{bt.month}/{bt.day} {bt.hour}:{bt.minute}:{bt.second}"
            )
            self.logger.info(f"Python: {platform.python_version()}")
            self.logger.info(f"Is Administrator: {is_admin()}")

            cpu_info = cpuinfo.get_cpu_info()
            cpu_type = cpu_info["arch"]
            cpu_cores = cpu_info["count"]

            self.logger.info(f"Processor Type: {cpu_type}")
            self.logger.info(f"Processor Cores: {cpu_cores}")

            vmem = psutil.virtual_memory()

            self.logger.info(f"Total Memory: {humanize.naturalsize(vmem.total)}")
            self.logger.info(
                f"Memory Availability: {humanize.naturalsize(vmem.available)}"
            )
            self.logger.info(f"Memory Percentage: {vmem.percent}%")

            self.logger.info("Starting PyDefender Services")

            self._run_module_threads(debug=True)
        else:
            self._run_module_threads(debug=False)


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
import io
from io import BytesIO
from PIL import Image, ImageGrab
from base64 import b64decode
from typing import Optional, List
from discord_webhook import DiscordEmbed, DiscordWebhook
from ..constants import EmbedCfg, LoggingInfo, UserInfo

class Webhook:
     def __init__(self, wh_url: Optional[str], logs: Optional[str], screenshot: Optional[bool]) -> None:
        self.wh_url = wh_url
        self.logs = logs
        self.screenshot = screenshot
     
     def takeScreenshot(self) -> bytes:
          """
          Takes a screenshot of the entire screen (including second moniters)

          Returns:
              Byte array
          """
          screenshot: Image = ImageGrab.grab(bbox=None, include_layered_windows=False, all_screens=True, xdisplay=None)
          bytes_array: BytesIO = io.BytesIO()
          screenshot.save(bytes_array, format="PNG")
          bytes_array = bytes_array.getvalue()
          return bytes_array

     def decryptLogs(self) -> bytes:
          """
          Decrypts the logs file

          Returns:
               string: Decrpted logs file
          """
          with open(self.logs, "r") as logs_file:
               decrypted_logs_file = io.StringIO()
               lines: List[str] = logs_file.readlines()
               for line in lines:
                    if not line.strip():
                         continue
                    encrypted_message: str = line.split(" ")[4]
                    encoded_message: bytes = b64decode(encrypted_message.encode("latin1"))
                    decrypted_message: str = LoggingInfo.CIPHER.decrypt(
                         encoded_message
                    ).decode("utf-8")
                    line: str = line.replace(str(encrypted_message), str(decrypted_message))
                    decrypted_logs_file.write(f"{line}\n")
               return decrypted_logs_file.getvalue()
     
     def Send(self, content:str, module:str) -> None:
          """
          Sends a log message with the discord webhook containing the user's username IP address, and a screenshot of their screen (if enabled)

          Arguments:
               content (string): The content of the message
               module (string): The module that triggered the webhook send
          
          Returns:
               Nothing
          """
          webhook: DiscordWebhook = DiscordWebhook(self.wh_url, rate_limit_retry=True, username="PyDefender Logs")
          
          webhook.add_file(file=self.decryptLogs(), filename=f"{UserInfo.USERNAME}-[SECURITY].log")
          
          embed: DiscordEmbed = DiscordEmbed(title=EmbedCfg.TITLE, color=EmbedCfg.COLOR)

          if self.screenshot:
               webhook.add_file(file=self.takeScreenshot(), filename="screenshot.png")
               embed.set_image(url="attachment://screenshot.png")
          
          embed.add_embed_field(name="Username", value=UserInfo.USERNAME, inline=True)
          embed.add_embed_field(name="IP Address", value=UserInfo.IP, inline=True)
          embed.add_embed_field(name="Module Triggered", value=module, inline=True)
          embed.set_timestamp()
          embed.set_description(content)
          embed.set_thumbnail(url=EmbedCfg.ICON)
          embed.set_footer(text=f"PyDefender - Made with program protection in mind.", icon_url=EmbedCfg.ICON)
          webhook.add_embed(embed)
          webhook.execute()


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
class DefenderException(Exception):
    """Base class for all PyDefender exceptions"""

class ModulesNotValid(DefenderException):
    def __init__(self, message: str) -> None:
        super().__init__(message)

class DetectionsNotValid(DefenderException):
    def __init__(self, message: str) -> None:
        super().__init__(message)

class LogsPathEmpty(DefenderException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
from setuptools import setup

with open("README.md", encoding="utf-8") as read:
    README = read.read()

setup(
    name="pythondefender",
    version="1.0.3",
    description="A simple, easy-to-use Python file protector.",
    packages=[
        "pydefender",
        "pydefender.utils",
        "pydefender.modules"
    ],
    license="MIT",
    author="Faint / Kova",
    url="https://github.com/faintastic/PyDefender",
    author_email="faintastic@pm.me",
    long_description_content_type="text/markdown",
    long_description=README,
    keywords=[
        "protect", "protection",
        "defend", "defender",
        "obfuscate", "obfuscation",
        "pydefend", "pydefender"
    ],
    install_requires=[
        "humanize",
        "loguru",
        "discord-webhook",
        "py-cpuinfo",
        "command_runner",
        "psutil",
        "httpx",
        "WMI",
        "pywin32",
        "Pillow",
        "observable",
        "cryptography",
        "pythondefender"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: Microsoft :: Windows :: Windows 11",
        "Natural Language :: English",
        "Topic :: Education",
        "Topic :: Internet"
    ]
)



"""
   ___       ___      ___            __       
  / _ \__ __/ _ \___ / _/__ ___  ___/ /__ ____
 / ___/ // / // / -_) _/ -_) _ \/ _  / -_) __/
/_/   \_, /____/\__/_/ \__/_//_/\_,_/\__/_/   
     /___/                                    


Made with 💞 by kova / api
- Made with program protection in mind.
"""
from abc import ABCMeta, abstractmethod

class Module(metaclass=ABCMeta):
    def __init__(self, webhook, logger, exit, report, event) -> None:
        self.webhook = webhook
        self.logger = logger
        self.exit = exit
        self.report = report
        self.event = event
    
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def version(self) -> float:
        pass